// Generated by CoffeeScript 2.5.1
var LanguageServerPlugin, baseTheme, documentUri, formatContents, languageId, offsetToPos, posToOffset, prefixMatch, rootUri, serverUri, toSet;

import {
  RequestManager,
  Client,
  WebSocketTransport
} from '@open-rpc/client-js';

import {
  Facet
} from "@codemirror/state";

import {
  EditorView,
  ViewPlugin,
  themeClass
} from "@codemirror/view";

import {
  linter
} from '@codemirror/lint';

import {
  showTooltip,
  hoverTooltip
} from '@codemirror/tooltip';

import {
  autocompletion
} from '@codemirror/autocomplete';

import {
  Deferred
} from 'promise.coffee';

serverUri = Facet.define({
  combine: (values) => {
    return values.reduce(((_, v) => {
      return v;
    }), '');
  }
});

rootUri = Facet.define({
  combine: (values) => {
    return values.reduce(((_, v) => {
      return v;
    }), '');
  }
});

documentUri = Facet.define({
  combine: (values) => {
    return values.reduce(((_, v) => {
      return v;
    }), '');
  }
});

languageId = Facet.define({
  combine: (values) => {
    return values.reduce(((_, v) => {
      return v;
    }), '');
  }
});

LanguageServerPlugin = class LanguageServerPlugin {
  constructor(view1) {
    this.view = view1;
    this.rootUri = this.view.state.facet(rootUri);
    this.documentUri = this.view.state.facet(documentUri);
    this.languageId = this.view.state.facet(languageId);
    this.documentVersion = 0;
    this.transport = new WebSocketTransport(this.view.state.facet(serverUri));
    this.requestManager = new RequestManager([this.transport]);
    this.client = new Client(this.requestManager);
    this.client.onNotification(() => {
      return this.processNotification(...arguments);
    });
    this.initialize({
      documentText: this.view.state.doc.toString()
    });
  }

  update({view, state, docChanged}) {
    return null;
  }

  destroy() {
    return this.client.close();
  }

  initialize({documentText}) {
    var ref;
    return (ref = this.client) != null ? ref.request({
      method: 'initialize',
      params: {
        capabilities: {
          textDocument: {
            hover: {
              dynamicRegistration: true,
              contentFormat: ['plaintext', 'markdown']
            },
            synchronization: {
              dynamicRegistration: true,
              willSave: false,
              didSave: false,
              willSaveWaitUntil: false
            },
            completion: {
              dynamicRegistration: true,
              completionItem: {
                snippetSupport: false,
                commitCharactersSupport: true,
                documentationFormat: ['plaintext', 'markdown'],
                deprecatedSupport: false,
                preselectSupport: false
              },
              contextSupport: false
            },
            signatureHelp: {
              dynamicRegistration: true,
              signatureInformation: {
                documentationFormat: ['plaintext', 'markdown']
              }
            },
            declaration: {
              dynamicRegistration: true,
              linkSupport: true
            },
            definition: {
              dynamicRegistration: true,
              linkSupport: true
            },
            typeDefinition: {
              dynamicRegistration: true,
              linkSupport: true
            },
            implementation: {
              dynamicRegistration: true,
              linkSupport: true
            }
          },
          workspace: {
            didChangeConfiguration: {
              dynamicRegistration: true
            }
          }
        },
        initializationOptions: null,
        processId: null,
        rootUri: this.rootUri,
        workspaceFolders: [
          {
            name: 'root',
            uri: this.rootUri
          }
        ]
      }
    }).then(({capabilities, serverInfo}) => {
      var ref1, ref2;
      this.capabilities = capabilities;
      if ((ref1 = this.client) != null) {
        ref1.notify({
          method: 'initialized',
          params: {}
        });
      }
      if ((ref2 = this.client) != null) {
        ref2.notify({
          method: 'textDocument/didOpen',
          params: {
            textDocument: {
              uri: this.documentUri,
              languageId: this.languageId,
              text: documentText,
              version: this.documentVersion
            }
          }
        });
      }
      this.ready = true;
      if (this.sendOnReady) {
        this.sendOnReady = false;
        return this.sendChange({
          documentText: documentText
        });
      }
    }) : void 0;
  }

  sendChange({documentText}) {
    var e, ref;
    if (!this.ready) {
      this.sendOnReady = true;
      return;
    }
    try {
      return (ref = this.client) != null ? ref.notify({
        method: 'textDocument/didChange',
        params: {
          textDocument: {
            uri: this.documentUri,
            version: this.documentVersion++
          },
          contentChanges: [
            {
              text: documentText
            }
          ]
        }
      }) : void 0;
    } catch (error) {
      e = error;
      return console.error(e);
    }
  }

  requestDiagnostics(view) {
    var deffered, ref;
    this.diagnosticsVersion = this.documentVersion;
    this.sendChange({
      documentText: view.state.doc.toString()
    });
    deffered = new Deferred();
    if ((ref = this.diagnosticsDeferred) != null) {
      ref.resolve([]);
    }
    this.diagnosticsDeferred = deffered;
    return deffered.promise;
  }

  requestHoverTooltip(view, {line, character}) {
    var ref;
    if (!this.ready || !this.capabilities.hoverProvider) {
      return null;
    }
    this.sendChange({
      documentText: view.state.doc.toString()
    });
    if ((ref = this.client) != null) {
      ref.request({
        method: 'textDocument/hover',
        params: {
          textDocument: {
            uri: this.documentUri
          },
          position: {
            line: line,
            character: character
          }
        }
      }).then((result) => {
        var contents, end, pos, range;
        if (!result) {
          this.view.dispatch({
            reconfigure: {
              hoverTooltip: []
            }
          });
          return;
        }
        ({contents, range} = result);
        pos = posToOffset(view.state.doc, {
          line: line,
          character: character
        });
        end = void 0;
        if (range) {
          pos = posToOffset(view.state.doc, range.start);
          end = posToOffset(view.state.doc, range.end);
        }
        this.view.dispatch({
          reconfigure: {
            hoverTooltip: showTooltip.of({
              pos: pos,
              end: end,
              create: function(view) {
                var el;
                el = document.createElement('div');
                el.textContent = formatContents(contents);
                return {
                  dom: el
                };
              },
              style: 'documentation',
              above: true
            })
          }
        });
      });
    }
    return null;
  }

  requestCompletion(context, {line, character}, {triggerKind, triggerCharacter}) {
    var ref;
    if (!this.ready || !this.capabilities.completionProvider) {
      return null;
    }
    this.sendChange({
      documentText: context.state.doc.toString()
    });
    return (ref = this.client) != null ? ref.request({
      method: 'textDocument/completion',
      params: {
        textDocument: {
          uri: this.documentUri
        },
        position: {
          line: line,
          character: character
        },
        context: {
          triggerKind: triggerKind,
          triggerCharacter: triggerCharacter
        }
      }
    }).then((result) => {
      var match, options, pos, span, token, word;
      if (!result) {
        return null;
      }
      options = result.items.map(({detail, label, kind, textEdit, documentation, sortText, filterText}) => {
        var completion;
        completion = {
          label: label,
          detail: detail,
          apply: label,
          type: {
            1: 'text',
            2: 'method',
            3: 'function',
            // 4: Constructor
            // 5: Field
            6: 'variable',
            7: 'class',
            8: 'interface',
            // 9: Module
            10: 'property',
            // 11: Unit
            // 12: Value
            13: 'enum',
            14: 'keyword',
            // 15: Snippet
            // 16: Color
            // 17: File
            // 18: Reference
            // 19: Folder
            // 20: EnumMember
            21: 'constant'
          }[kind],
          sortText: label,
          filterText: label
        };
        if (textEdit) {
          completion.apply = textEdit.newText;
        }
        if (documentation) {
          completion.info = formatContents(documentation);
        }
        if (sortText) {
          completion.sortText = sortText;
        }
        if (filterText) {
          completion.filterText = filterText;
        }
        return completion;
      });
      [span, match] = prefixMatch(options);
      token = context.matchBefore(match);
      pos = context.pos;
      if (token) {
        pos = token.from;
        word = token.text.toLowerCase();
        if (/^\w+$/.test(word)) {
          options = options.filter(({filterText}) => {
            return filterText.toLowerCase().indexOf(word) === 0;
          });
          options = options.sort((a, b) => {
            switch (true) {
              case a.apply.indexOf(token.text) === 0 && b.apply.indexOf(token.text) !== 0:
                return -1;
              case a.apply.indexOf(token.text) !== 0 && b.apply.indexOf(token.text) === 0:
                return 1;
              default:
                return 0;
            }
          });
        }
      }
      return {
        'from': pos,
        options: options
      };
    }) : void 0;
  }

  processNotification({method}) {
    var e;
    try {
      switch (method) {
        case 'textDocument/publishDiagnostics':
          this.processDiagnostics(...arguments);
      }
    } catch (error) {
      e = error;
      return console.error(e);
    }
  }

  processDiagnostics({params}) {
    var annotations, ref;
    if (this.documentVersion > this.diagnosticsVersion + 1) {
      return;
    }
    annotations = params.diagnostics.map(({range, message, severity}) => {
      return {
        'from': posToOffset(this.view.state.doc, range.start),
        to: posToOffset(this.view.state.doc, range.end),
        severity: {
          1: 'error',
          2: 'warning',
          3: 'info',
          4: 'info'
        }[severity],
        message: message
      };
    });
    annotations.sort((a, b) => {
      switch (true) {
        case a.from < b.from:
          return -1;
        case a.from > b.from:
          return 1;
        default:
          return 0;
      }
    });
    return (ref = this.diagnosticsDeferred) != null ? ref.resolve(annotations) : void 0;
  }

  clearTooltip() {
    return this.view.dispatch({
      reconfigure: {
        hoverTooltip: []
      }
    });
  }

};

export var languageServer = function(options) {
  var plugin;
  plugin = null;
  return [
    serverUri.of(options.serverUri),
    rootUri.of(options.rootUri),
    documentUri.of(options.documentUri),
    languageId.of(options.languageId),
    ViewPlugin.define((view) => {
      plugin = new LanguageServerPlugin(view);
      return plugin;
    },
    {
      eventHandlers: {
        keyup: function() {
          return this.clearTooltip();
        },
        click: function() {
          return this.clearTooltip();
        }
      }
    }),
    linter((view) => {
      return plugin != null ? plugin.requestDiagnostics(view) : void 0;
    }),
    hoverTooltip((view,
    pos,
    side) => {
      return plugin != null ? plugin.requestHoverTooltip(view,
    offsetToPos(view.state.doc,
    pos)) : void 0;
    }),
    autocompletion({
      override: [
        (context) => {
          var explicit,
        line,
        pos,
        ref,
        ref1,
        ref2,
        state,
        trigChar,
        trigKind;
          if (plugin == null) {
            return null;
          }
          ({state,
        pos,
        explicit} = context);
          line = state.doc.lineAt(pos);
          trigKind = 1; // Invoked
          trigChar = void 0;
          if (!explicit && ((ref = plugin.capabilities) != null ? (ref1 = ref.completionProvider) != null ? (ref2 = ref1.triggerCharacters) != null ? ref2.indexOf(line.text[pos - line.from - 1]) : void 0 : void 0 : void 0) >= 0) {
            trigKind = 2; // TriggerCharacter
            trigChar = line.text[pos - line.from - 1];
          }
          if (trigKind === 1 && !context.matchBefore(/\w+$/)) {
            return null;
          }
          return plugin.requestCompletion(context,
        offsetToPos(state.doc,
        pos),
        {
            triggerKind: trigKind,
            triggerCharacter: trigChar
          });
        }
      ]
    }),
    baseTheme
  ];
};

posToOffset = function(doc, pos) {
  return doc.line(pos.line + 1).from + pos.character;
};

offsetToPos = function(doc, offset) {
  var line;
  line = doc.lineAt(offset);
  return {
    line: line.number - 1,
    character: offset - line.from
  };
};

formatContents = function(contents) {
  var c, j, len, text;
  if (Array.isArray(contents)) {
    text = '';
    for (j = 0, len = contents.length; j < len; j++) {
      c = contents[j];
      text += formatContents(c) + '\n\n';
    }
    return text;
  } else if (typeof contents === 'string') {
    return contents;
  } else {
    return contents.value;
  }
};

toSet = function(chars) {
  var flat, preamble, words;
  preamble = "";
  flat = Object.keys(chars).join('');
  words = /\w/.test(flat);
  if (words) {
    preamble += "\\w";
    flat = flat.replace(/\w/g, '');
  }
  return `[${preamble}${flat.replace(/[^\w\s]/g, '\\$&')}]`;
};

prefixMatch = function(options) {
  var apply, first, i, j, k, len, ref, rest, source;
  first = {};
  rest = {};
  for (j = 0, len = options.length; j < len; j++) {
    ({apply} = options[j]);
    first[apply[0]] = true;
    for (i = k = 1, ref = apply.length; (1 <= ref ? k < ref : k > ref); i = 1 <= ref ? ++k : --k) {
      rest[apply[i]] = true;
    }
  }
  source = toSet(first) + toSet(rest) + "*$";
  return [new RegExp("^" + source), new RegExp(source)];
};

baseTheme = EditorView.baseTheme({
  '$tooltip.documentation': {
    display: 'block',
    marginLeft: '0',
    padding: '3px 6px 3px 8px',
    borderLeft: '5px solid #999',
    whiteSpace: 'pre'
  },
  '$tooltip.lint': {
    whiteSpace: 'pre'
  }
});
